##  (c) 2004 Shane T. Mueller, Ph.D.
##  smueller@obereed.net
##  stmuelle@indiana.edu
##
##  This routines compare standard R trees using metrics described in 
##  Boorman, S. A., and Olivier, D. C. (1973). Metrics on Finite Trees. 
##  Journal of Mathematical Psychology, 10, 26-59.
##  It requires that a tree be described by the the order, merge, and
##  height data.


## This takes a tree and represents it as a nested list of subsets.
## this can be done directly from the $merge data structure.
MakeSets <- function(tree)
{
  ## There will be a total number of sets equal to the length of tree$merge,
  ##, so make an initial set structure
  sets <- as.list(rep(0,nrow(tree$merge)))

  for(i in 1:nrow(tree$merge))
    {
      terminals <- tree$merge[i,]<0

      ##Now, add any terminals to the tmpset.
      tmpSet <- as.list(-tree$merge[i,terminals])

      ##Now, for any non-terminals, add the elements
      ##from the node that it includes.
     for(j in tree$merge[i,!terminals])
       {
         tmpSet <- append(tmpSet,sets[[j]])
       }
      sets[[i]] <- tmpSet
    }

  sets
}

## This is a different way of partitioning a tree, as found on
## page 32.  It uses MakeSets in its calculation.
## It operates by creating partitions of the data set,
## by "slicing" it at each node, with all more detailed
## tree structure being kept as distinct partitions and all
## less detailed partitions being ignored.  For example,
## the lowest partition will contain all elements in a single partition.
## The next highest will contain two sets, then three, etc.

##This amounts to finding the sequence of 'cut-points', dividing
##the current partition in two. This is just order(height)
GetPartition <- function(tree, cuts)
  {
    #splits <- sort(order(-tree$height)[1:(level-1)])
    indexes <- 1:(length(tree$height))
    splits <- (indexes[rank(-tree$height) < cuts])
    splits <- append(splits,length(tree$height)+1)
  
    
    ##OK, we now now the places where tree$order is split.
    ##It is sorted for convenience.


    lowerbound <- 0
    returnlist <- as.list(rep(0,length(splits)))
    for(i in 1:length(splits))
      {
    
        upperbound <- splits[i]
        returnlist[[i]] <- tree$order[lowerbound:upperbound]
        lowerbound <- upperbound+1
      }
    returnlist
  }


##This creates a partition by specifying a certain level.
##This only accepts groups that are less than or equal to the level
##specified.
GetPartitionAtLevel <- function(tree, heights, level)
  {
    ##heights is a new set of heights applied to the tree--
    ##a stand-in for tree$height.
    
    indexes <- 1:(length(heights))
    splits <- (indexes[heights  > level])
    splits <- append(splits, length(heights)+1)

    ##OK, we now now the places where tree$order is split.
    ##It is sorted for convenience.
        
    lowerbound <- 0
    returnlist <- as.list(rep(0,length(splits)))


    if(length(splits) > 0)
      {
        for(i in 1:length(splits))
          {
            upperbound <- splits[i]
            returnlist[[i]] <- tree$order[lowerbound:upperbound]
            lowerbound <- upperbound+1
          }
      }

    returnlist
    
  }

##This applies GetPartition iteratively to construct
##the entire partition chain.
MakePartitionChain <- function(tree)
  {
    chain <- NULL
    for(i in 1:1:(nrow(tree$merge)+1))
      {
        chain <- append(chain,list(GetPartition(tree,i)))
      }
    rev(chain)
  }

### This makes the partition chain at specified levels, instead of
### at every integer. heights is an alternate for tree$height.
MakePartitionChainAtLevels <- function(tree, heights, levels)
  {
    chain <- NULL
    for(i in levels)
      {
        chain <- append(chain,list(GetPartitionAtLevel(tree,heights, i)))
      }
    chain
  }

## This finds the size of the union of two partitions.
## The union is defined, basically, as the items must
## be in the same cell in one of the two partitions.
PartitionUnion <- function(part1, part2)
  {

    ##Start with partition 1.  The union will be
    ##no bigger than that partition. Combine any
    ##groups 

    part1size <- length(part1)
    part2size <- length(part2)

    ##Create 'used' vectors telling whether
    ##each element in part1 and part2 have been used.
    unused1 <- rep(T,part1size)
    unused2 <- rep(T,part2size)
    
    unionPart <- NULL
    
    ##Iterate through the groups in partition 1
    for(i in 1:part1size)
      {
        ##Set flags that tell whether any partitions were merged.
        merged1 <- T
        merged2 <- T
        
        ##Go through all the unused cells in partition 1
        if(unused1[i] == T)
          {
            newPartition <- part1[[i]]
            unused1[i] <- F
        
            ##Now that I have the core of a new partition,
            ##Go through all the partitions of part2.  Add any
            ##cells that have non-null unions with newPartition.

            cycle <- 1
                
            ##We should cycle until we get two cycles in a row where nothing is merged.
            while(merged1==T || merged2==T)
              {
                
                ##alternate between scanning partition 1 and partition 2
                if(cycle %% 2)          ##On odd cycles, scan partition 2
                  {
                    merged2 <- F       #Nothing has been merged yet.

                    for(j in 1:part2size)
                      {
                        if(unused2[j]==T)  ##Only examine it if the cell hasn't been used yet
                          {
                            if(length(intersect(newPartition,part2[[j]]))!=0)
                              {

                                newPartition <- union( newPartition,part2[[j]])
                                unused2[j] <- F
                                merged2 <- T
                              }
                          }
                      }
                    
                  }
                else                      ##on even cycles, scan partition 1
                  {
                        merged1 <- F       #Nothing has been merged yet.
                        for(j in 1:part1size)
                          {
                            if(unused1[j]==T)  ##Only examine it if the cell hasn't been used yet
                              {
                                if(length(intersect(newPartition,part1[[j]]))!=0)
                                  {
                                    newPartition <- union( newPartition,part1[[j]])
                                    unused1[j] <- F
                                    merged1 <- T
                                  }
                              }
                          }
                      }
                
                cycle <-  cycle + 1
              }

            ##Now, newPartition represents a new union. Add it to the union lists.
            unionPart <- append(unionPart,list(newPartition))
          }
        
      }
    unionPart
  }

## This finds the intersection of two partitions.
## Two items are in the same set in the intersection if
## they are together in both of the partitions.
PartitionIntersection <- function(part1,part2)
  {

    ##To do this, we find all of the non-null intersections.
    intersectPart <- NULL
    
    for(i in part1)
      {
        for(j in part2)
          {
            inter <- intersect(i,j)
            if(length(inter)!=0)
              {
                ##inter is a valid intersection.  Add it to the return list.
                intersectPart <- append(intersectPart,list(inter))
              }
          }
      }
    
    ##Now, all partitions have been compared.
    ##Add to the running count the number of unused symbols,
    ##each of which are their own group.
    intersectPart
  }


##This calculates the "C" distance between two partitions, as
##defined by Boorman & Olivier, Def. 1.7, p. 33.
CDistance <- function(part1, part2)
  {
    length(part1) + length(part2) -
      2 * length(PartitionUnion(part1,part2))
  }


##normchoose2 is used in DDistance.  It goes through
##each cell of a partition, finds the size of that value choose 2,
##returning the sum.
SumNormChoose2 <- function(x)
  {
    total <- 0

    for(i in x)
      {
        norm <- length(i)
        total <- total + norm * (norm-1)/2
      }

    total
  }



##This calculates the "D" distance between two partitions, as
##defined by Boorman & Olivier, Def. 1.9, p. 33.
DDistance <- function(part1,part2)
  {
    SumNormChoose2(part1) +  SumNormChoose2(part2) - 2 *  SumNormChoose2(PartitionIntersection(part1,part2))
  }


## This returns the correct partition for
## a specified height.  This means the index of
## the number of the next larger partition than the given height.
GetPartIndex <- function(height,heights)
{
  ##we must return at least 1
  max(length(heights)-sum((height<=heights)),1)
}



##This is a generic 'C' distance
##This takes two partitions with two sets of 'heights',
##indicating the numeric boundaries between partitions on the
##lattice.
treedist.C<- function(part1,  part2, partheight1, partheight2)
  {
    ##Now, we must take the integral over the heights
    allheights <- sort(unique(c(0,partheight1,partheight2)))
        
    ##Now, move through each height from the lowest to the highest,
    ##At each stage calculating the appropriate inter-partition distance,
    ##multiplying it by the delta, and adding it to the sum.

    total <- 0

    for(i in 1:(length(allheights)-1))
      {
        ##Determine the lower and upper bounds of comparison
        lower <- allheights[i]
        upper <- allheights[i+1]
        delta <- upper-lower

        ##Extract the appropriate partition from the lattice
        part1index <- GetPartIndex(upper, partheight1)
        part2index <- GetPartIndex(upper, partheight2)

        ##Calculate distance between the two partitions.
        distance <- CDistance(part1[[part1index]],part2[[part2index]])
        total <- total + delta * distance

      }

    total
  }



##This is a generic 'D' distance
treedist.D <- function(part1,  part2, partheight1, partheight2)
  {
    ##Now, we must take the integral over the heights
    allheights <- sort(unique(c(partheight1,partheight2)))
    
    ##Now, move through each height from the lowest to the highest,
    ##At each stage calculating the appropriate inter-partition distance,
    ##multiplying it by the delta, and adding it to the sum.
    
    total <- 0
    
    for(i in 1:(length(allheights)-1))
      {

        ##Determine the lower and upper bounds of comparison
        lower <- allheights[i]
        upper <- allheights[i+1]

        delta <- upper-lower
        
        ##Extract the appropriate partition from the lattice
        part1index <- GetPartIndex(upper, partheight1)
        part2index <- GetPartIndex(upper, partheight2)

        ##Calculate distance between the two partitions.
        distance <- DDistance(part1[[part1index]],part2[[part2index]])

        #Increment the distance
        total <- total + delta * distance
      }

    total
  }

##This calculates the 'valued' C distance between two trees
## (accounts for different heights of branches)
treedist.mC <- function(tree1, tree2)
{
  ##For valued trees, heights are taken directly from the 
  ## heights of the tree objects.

  height1 <- tree1$height
  height2 <- tree2$height
  
  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- sort(c(0,unique(height1)))
  chainheight2 <- sort(c(0,unique(height2)))

  ##Produce the partition lattice for each tree at specified heights
  partchain1 <- MakePartitionChainAtLevels(tree1,height1,chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2,height2,chainheight2)

  ##Calculate the C distance between trees
  treedist.C(partchain1, partchain2, chainheight1, chainheight2)
}

##This calculates the 'valued' D distance between two trees
## (accounting for different heights of branches)
treedist.mD <- function(tree1, tree2)
{
  ##For valued trees, heights are taken directly from the 
  ## heights of the tree objects.
  height1 <- tree1$height
  height2 <- tree2$height

  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- sort(c(0,unique(height1)))
  chainheight2 <- sort(c(0,unique(height2)))

  ##Produce the partition lattice for each tree at specified heights 
  partchain1 <- MakePartitionChainAtLevels(tree1,height1,chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2,height2,chainheight2)

  ##Calculate the D distance between trees
  treedist.D(partchain1, partchain2, chainheight1, chainheight2)
}


##This calculates the 'ranked C distance between two trees
## (only considers rank of branch heights.)
##This amounts to the sum of the distances at each level.
treedist.rC <- function(tree1, tree2)
{

  ##For ranked trees, heights are taken from the
  ##ranked heights of the tree objects.
  height1 <- rank(tree1$height)
  height2 <- rank(tree2$height)

  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- c(0,sort(unique(height1)))
  chainheight2 <- c(0,sort(unique(height2)))

  ##Produce the partition lattice for each tree at specified heights 
  partchain1 <- MakePartitionChainAtLevels(tree1,height1,chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2,height2,chainheight2)

  ##Calculate the D distance between trees
  treedist.C(partchain1, partchain2, chainheight1, chainheight2)

}

##This calculates the 'ranked' D distance between two trees
## (only considers rank of branch heights.)
treedist.rD <- function(tree1, tree2)
{
  ##For ranked trees, heights are taken from the
  ##ranked heights of the tree objects.
  height1 <- rank(tree1$height)
  height2 <- rank(tree2$height)

  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- c(0,sort(unique(height1)))
  chainheight2 <- c(0,sort(unique(height2)))

  ##Produce the partition lattice for each tree at specified heights 
  partchain1 <- MakePartitionChainAtLevels(tree1,height1,chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2,height2,chainheight2)

  ##Calculate the D distance between trees
  treedist.D(partchain1, partchain2, chainheight1, chainheight2)
}


##This calculates tree distance for 'bare' trees, using the 'C' distance
treedist.bC <- function(tree1, tree2)
{

  ##For bare trees, we use the size of the sub-cluster as the height
  ##of the tree.  We can get this from finding the size of the
  ##MakeSets() output.
  ##sizes* is ordered from lowest branch to highest.
  sizes1 <-sapply( MakeSets(tree1),length)
  sizes2 <-sapply( MakeSets(tree2),length)
  
  
  ##height is ordered from left to right, replacing tree$height.
  height1 <- sizes1[order(order(tree1$height))]
  height2 <- sizes2[order(order(tree2$height))]

  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- c(0,sort(unique(height1)))
  chainheight2 <- c(0,sort(unique(height2)))

  ##Produce the partition lattice for each tree at specified heights 
  partchain1 <- MakePartitionChainAtLevels(tree1, height1, chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2, height2, chainheight2)

  ##Calculate the C distance between trees
  treedist.C(partchain1, partchain2,  chainheight1, chainheight2)
    
}

##This calculates tree distance for 'bare' trees, using the 'D' distance
treedist.bD <- function(tree1, tree2)
{
  ##For bare trees, we use the size of the sub-cluster as the height
  ##of the tree.  We can get this from finding the size of the
  ##MakeSets() output.
  ##sizes* is ordered from lowest branch to highest.
  sizes1 <-sapply( MakeSets(tree1),length)
  sizes2 <-sapply( MakeSets(tree2),length)
  
  
  ##height is ordered from left to right, replacing tree$height.
  height1 <- sizes1[order(order(tree1$height))]
  height2 <- sizes2[order(order(tree2$height))]

  ##Find the heights on which the partitions should be evaluated.
  ##e.g., only get one partition for ties.
  chainheight1 <- c(0,sort(unique(height1)))
  chainheight2 <- c(0,sort(unique(height2)))
  
  ##Produce the partition lattice for each tree at specified heights 
  partchain1 <- MakePartitionChainAtLevels(tree1, height1, chainheight1)
  partchain2 <- MakePartitionChainAtLevels(tree2, height2, chainheight2)
  

  ##Calculate the C distance between trees
  treedist.D(partchain1, partchain2,  chainheight1, chainheight2)

}




## This tests some of the sample data in Boorman and Olivier
## to determine whether the above procedures produce the same
## results.


##These are not *real* trees, but just carry enough information
##to be used by the above functions.
library(cluster)

labels <- c("Gf","Gm","Gs","Gd","F","M","So","D", "B", "Si", "U", "A", "Ne", "Ni", "C")

BO.example.KTM <- structure(list(
                                 order=1:15,
                                 merge=
                                 t(as.matrix(cbind(c(-5,-6),
                                       c(-9,-10),
                                       c(-7,-8),
                                       c(-11,-12),
                                       c(1,3),
                                       c(-1,-2),
                                       c(2,5),
                                       c(-13,-14),
                                       c(-3,6),
                                       c(4,8),
                                       c(-4,9),
                                       c(11,7),
                                       c(10,-15),
                                       c(12,13)))),
                            
                                 height=c(10.6, 24.7, 27.2, 29.5, 5.0, 10.3, 8.8, 18.1, 8.3, 51.5, 9.6, 25.8, 22.8, 37.2),
                                 order.lab=labels,
                                 call=c("BO", "X","sim"),
                                 ac = 0,
                                 method="average",
                                 diss=NULL
                                 ), class=c("agnes", "twins")
                            )

BO.example.KTF<- structure(list(
                                order=c(1,2,3,4,5,8,7,6,9,10,11,12,13,14,15),
                                merge=
                                t(as.matrix(cbind(
                                                  c(-5,-8),
                                                  c(-7,-6),
                                                  c(-1,-2),
                                                  c(-9,-10),
                                                  c(1,2),
                                                  c(-11,-12),
                                                  c(5,4),
                                                  c(-3,3),
                                                  c(-4,8),
                                                  c(-13,-14),
                                                  c(6,10),
                                                  c(9,7),
                                                  c(11,-15),
                                                  c(12,13)
                                                  ))),
                                
                                height=c(8.7, 18.8, 20.7, 31.6, 5.1, 10.8, 5.2, 16.1,
                                  9.0, 48.5, 12.5, 26.3, 23.5, 38.5),
                                
                                order.lab=labels[c(1,2,3,4,5,8,7,6,9,10,11,12,13,14,15)],
                                call=c("BO", "X","sim"),
                                ac = 0,
                                method="average",
                                diss=NULL
                                ), class=c("agnes", "twins")
                           )


BO.example.KLTM <- structure(list(
                                 order=1:15,
                                 merge=
                                  t(as.matrix(cbind(c(-5,-6),
                                                    c(-1,-2),
                                                    c(-7,-8),
                                                    c(-3,2),
                                                    c(1,3),
                                                    c(-4,4),
                                                    c(-9,-10),
                                                    c(-11,-12),
                                                    c(-13,-14),
                                                    c(5,6),
                                                    c(8,9),
                                                    c(7,11),
                                                    c(12,-15),
                                                    c(10,13)))),
                                  height=c(4.6, 8.9, 12.8, 23.7, 1.7, 10.1, 7.0, 43.7, 12.8, 34.7, 15.2, 27.1, 22.8, 39.8),
                                 order.lab=labels,
                                 call=c("BO", "X","sim"),
                                 ac = 0,
                                 method="average",
                                 diss=NULL
                                 ), class=c("agnes", "twins")
                            )

BO.example.KLTF <- structure(list(
                                order=c(1,2,4,3,6,5,8,7,9,10,14,12,13,11,15),
                                merge=
                                t(as.matrix(cbind(
                                                  c(-6,-5),
                                                  c(-8,1),
                                                  c(-1,-2),
                                                  c(-7,2),
                                                  c(-9,-10),
                                                  c(-4,3),
                                                  c(-13,-11),
                                                  c(-14,-12),
                                                  c(-3,6),
                                                  c(4,5),
                                                  c(8,7),
                                                  c(9,10),
                                                  c(11,-15),
                                                  c(12,13)
                                                  ))),
                                
                                height=c(5.6, 10.6, 15.4, 37.8, 1.6, 4.3, 8.8, 19.3, 9.9, 53.2, 14.4, 24.7, 12.7,38.2),
                                  order.lab=labels[c(1,2,4,3,6,5,8,7,9,10,14,12,13,11,15)],
                                call=c("BO", "X","sim"),
                                ac = 0,
                                method="average",
                                diss=NULL
                                ), class=c("agnes", "twins")
                           )


BO.example.KC <- structure(list(
                                 order=c(1,2,3,4,5,6,7,8,9,10,11,12,15,13,14),
                                 merge=
                                  t(as.matrix(cbind(c(-9,-10),
                                                    c(-5,-6),
                                                    c(-7,2),
                                                    c(-11,-12),
                                                    c(-1,-2), 
                                                    c(-8,3),  
                                                    c(-3,-4), 
                                                    c(-13,-14),
                                                    c(4,-15), 
                                                    c(1,6),    
                                                    c(5,7),    
                                                    c(8,9),    
                                                    c(10,11),  
                                                    c(12,13)))),
                                  height=c(19.7, 53.2, 31.3, 77.7, 12.3, 18.5, 29.1,47.7,10.2,94.2,19.4,47.1,71.8,37.0),
                                 order.lab=labels[c(1,2,3,4,5,6,7,8,9,10,11,12,15,13,14)],
                                 call=c("BO", "X","sim"),
                                 ac = 0,
                                 method="average",
                                 diss=NULL
                                 ), class=c("agnes", "twins")
                            )


##These are the actually distance matrices published by Boorman & Olivier:

mC.BO <- rbind(c(0,        0,     0, 0, 0),
               c(42.1,     0,     0, 0, 0),
               c(103.7, 96.4,     0, 0, 0),
               c(90.7,  80.4, 111.2, 0, 0),
               c(0,        0,     0, 0, 0)
               )

rC.BO <- rbind(
               c( 0,  0,  0,  0,  0),
               c(22,  0,  0,  0,  0),
               c(35, 33,  0,  0,  0),
               c(28, 26, 28,  0,  0),
               c(24, 32, 41, 36,  0)
               )
bC.BO <- rbind(
               c( 0,  0,  0,  0,  0),
               c( 4,  0,  0,  0,  0),
               c(22, 26,  0,  0,  0),
               c( 9, 11, 31,  0,  0),
               c( 9, 11, 31, 12,  0)
               )


mD.BO <- rbind(
               c( 0,          0,       0,  0,  0),
               c( 279.7,      0,       0,  0,  0),
               c( 1003.6, 861.4,       0,  0,  0),
               c( 435.1,  513.3,  1127.7,  0,  0),
               c( 0,          0,       0,  0,  0)
               )


rD.BO <- rbind(
               c( 0,     0,    0,   0,  0),
               c( 35,    0,    0,   0,  0),
               c( 158, 161,    0,   0,  0),
               c( 66,   58,  134,   0,  0),
               c( 87,   94,  159,  90,  0)
               )
